# coding=utf-8

import os
from . import SubProcess
from pathlib import Path

temp_comment = "temp_comment.txt"


class File_Log:
    def __init__(self, fileRev, author, date, timestamp, description):
        self.Revision = fileRev
        self.Author = author
        self.Date = date
        self.TimeStamp = int(timestamp.strip())
        self.Description = description
        self.commit_id = None

    def __repr__(self):
        return f"{self.Revision}\n{self.Author}\n{self.Date}\n{self.TimeStamp}\n{self.Description}"


class GitHelper():
    def __init__(self, remote_repo, workspace, encoding="utf-8"):
        self.remote_repo = remote_repo
        self.workspace = workspace
        self.version = "1.0"
        self.encoding = encoding

    def __run_command(self, command, cwd=None, active_info=True, silent=False, show_cmd=True):
        run_command = SubProcess.RunCMD()
        run_command.active_info = active_info
        run_command.silent = silent
        run_command.show_cmd = show_cmd
        run_command.encoding = self.encoding
        cwd = cwd if cwd else self.workspace

        rc, msg = run_command.run_command(command, cwd, skip_blank_lines=True)
        if rc == SubProcess.RC_CODE.RC_PASS:
            return True, msg
        else:
            return False, msg

    def git_list(self, dir_or_file=None):
        command = 'git ls-tree -r --name-only HEAD'
        if dir_or_file:
            result, output = self.__run_command(command, cwd=dir_or_file)
        else:
            result, output = self.__run_command(command)
        if result:
            print("List successful!")
            return output
        else:
            print("List failed!")
            return None

    def git_fetch_all(self):
        command = "git fetch --all"
        result, output = self.__run_command(command)
        if result:
            return True
        else:
            return False

    def git_get_local_version(self):
        command = "git rev-parse HEAD"
        result, output = self.__run_command(command)
        if result:
            return "".join(output).strip()
        else:
            return None

    def git_get_remote_version(self):
        if self.git_fetch_all():
            command = "git rev-parse origin/HEAD"
            result, output = self.__run_command(command)
            if result:
                return "".join(output).strip()
            else:
                return None
        else:
            return None

    def git_get_local_commit_date(self, ):
        logs = self.git_log(path=None, num=1, more_args="HEAD", cwd=self.workspace)
        if logs:
            return logs[0].TimeStamp
        return None

    def git_get_remote_commit_date(self):
        if self.git_fetch_all():
            logs = self.git_log(path=None, num=1, more_args="origin/HEAD", cwd=self.workspace)
            if logs:
                return logs[0].TimeStamp
        return None

    def git_get_actions(self, first_run=False):
        """
        Added (A)
        Copied (C)
        Deleted (D)
        Modified (M)
        Renamed (R)
        changed (T)
        are Unmerged (U)
        are Unknown (X)
        or have had their pairing Broken (B).
         """
        new_actions = []
        if not os.path.exists(self.workspace):
            first_run = True
        if first_run:
            self.git_clone()
            initial_version = self.git_get_initial_revision()
            local_version = self.git_get_local_version()
            actions = self.git_diff(local_version, initial_version)
        else:
            local_version = self.git_get_local_version()
            remote_version = self.git_get_remote_version()
            if local_version == remote_version:
                return new_actions  # 如果版本一致，则actions为空
            else:
                actions = self.git_diff(remote_version, local_version)
        for action in actions:
            if not (action.strip().startswith("M")
                    or action.strip().startswith("A")
                    or action.strip().startswith("D")
                    or action.strip().startswith("R")):
                continue
            action_act = action.strip().split()[0]
            action_path = action.strip()[len(action_act):].strip()
            if action.strip().startswith("R"):
                rm_file, new_file = [x.strip() for x in action_path.split('\t')]
                new_actions.append(("D", rm_file))
                new_actions.append(("A", new_file))
            else:
                new_actions.append((action_act, action_path))
        return new_actions

    def git_clone(self):
        # clone
        command = f'git clone "{self.remote_repo}" "{self.workspace}"'
        result, output = self.__run_command(command, cwd=os.getcwd(), active_info=True)
        print("".join(output))
        if result:
            return True
        else:
            return False

    def git_merge(self):
        command = f'git merge origin/main'
        result, output = self.__run_command(command, active_info=True)
        print("".join(output))
        if result:
            return True
        else:
            return False

    def git_pull(self):
        command = f'git pull'
        result, output = self.__run_command(command, active_info=True)
        print("".join(output))
        if result:
            return True
        else:
            return False

    def git_get_initial_revision(self):
        command = f"git log --oneline --pretty=%H"
        result, output = self.__run_command(command)
        revisions = [x for x in output if x.strip()]
        if result:
            return revisions[-1].strip()
        else:
            return None

    def git_diff(self, local_version, remote_version):
        diff_actions = []
        pre_set = " git config core.quotepath false"
        command = f"{pre_set} & git diff {remote_version} {local_version} --name-status"
        result, output = self.__run_command(command)
        if result:
            print("Get diff succeed!")
            diff_actions = output
        else:
            print("Get diff failed!")
        return diff_actions

    def git_add(self, path):
        path = Path(path)
        if path.exists():
            command = f'git add "{path}"'
            result, output = self.__run_command(command)
            if result:
                print("Git add successful!")
                return True
            else:
                print("Git add failed!")
                return False
        else:
            return False

    def git_rm(self, path):
        path = Path(path)
        if path.exists():
            command = f'git rm "{path}"'
            result, output = self.__run_command(command)
            if result:
                print("Git rm successful!")
                return True
            else:
                print("Git rm failed!")
                return False
        else:
            return False

    def git_commit(self, path="", comment=""):
        comment_file = self.prepare_comment_file(comment)
        try:
            if path:
                command = f'git commit "{path}" -F "{comment_file}"'
            else:
                command = f'git commit -F "{comment_file}"'
            result, output = self.__run_command(command)
            if result:
                for line in output:
                    print(line)
                print("Commit successful!")
            else:
                for line in output:
                    print(line)
                print("Commit failed!")
            return result
        finally:
            if os.path.exists(comment_file):
                os.remove(comment_file)

    def git_push(self):
        command = f"git push"
        result, output = self.__run_command(command)
        if result:
            print("Git push successful!")
            return True
        else:
            print("Git push failed!")
            print('\n'.join(output))
            return False

    def git_add_commit_push(self, path, comment):
        if self.git_add(path):
            if self.git_commit(path, comment):
                if self.git_push():
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False

    def git_head_revision(self):
        command = "git rev-parse HEAD"
        result, output = self.__run_command(command)
        if result:
            print("Git get head revision successful!")
            return output
        else:
            print("Git get head revision failed!")
            return None

    def git_log(self, path=None, num=100, more_args="", cwd=None):
        logResult = []
        split = "***********************LOGEND*********************"
        if path:
            cmd = f'git log -n {num} --pretty="Revision:%H%nAuthor:%ae%nDate:%ci%nTimeStamp:%ct%nDescription:%B%n{split}" -- "{path}"'
        else:
            cmd = f'git log -n {num} --pretty="Revision:%H%nAuthor:%ae%nDate:%ci%nTimeStamp:%ct%nDescription:%B%n{split}" {more_args}'
        result, output = self.__run_command(cmd, cwd=cwd)
        if result:
            revision = ""
            author = ""
            date = ""
            timestamp = ""
            description = ""
            description_start = False
            for line in output:
                line_text = line.strip()
                if line_text.startswith("Revision:"):
                    revision = line_text[9:]
                elif line_text.startswith("Author:"):
                    author = line_text[7:]
                elif line_text.startswith("Date:"):
                    date = line_text[5:]
                elif line_text.startswith("TimeStamp:"):
                    timestamp = line_text[10:]
                elif line_text.startswith("Description:"):
                    description_start = True
                    description += line.lstrip()[12:]
                elif line_text == split:
                    file_log = File_Log(revision, author, date, timestamp, description.strip())
                    logResult.append(file_log)
                    description_start = False
                    description = ""
                else:
                    if description_start:
                        description += line
        return logResult

    @staticmethod
    def prepare_comment_file(comment):
        with open(temp_comment, "w", encoding="utf-8") as f:
            f.write(comment)
        return os.path.realpath(temp_comment)



# coding=utf-8

from copy import copy
from openpyxl.utils import get_column_letter
from openpyxl.styles.colors import COLOR_INDEX
from openpyxl.styles import Border, Side, PatternFill, Alignment
from enum import Enum
from pathlib import Path

MERGED_VALUE = "merged"


class Color(Enum):
    BLACK = COLOR_INDEX[0]
    WHITE = COLOR_INDEX[1]
    RED = COLOR_INDEX[2]
    GREEN = COLOR_INDEX[3]
    BLUE = COLOR_INDEX[4]
    YELLOW = COLOR_INDEX[5]
    PINK = COLOR_INDEX[6]
    LIGHT_BLUE = COLOR_INDEX[7]
    GRAY = COLOR_INDEX[22]
    ORANGE = COLOR_INDEX[52]


class Fill(Enum):
    BLACK = PatternFill("solid", fgColor=Color.BLACK.value)
    WHITE = PatternFill("solid", fgColor=Color.WHITE.value)
    RED = PatternFill("solid", fgColor=Color.RED.value)
    GREEN = PatternFill("solid", fgColor=Color.GREEN.value)
    BLUE = PatternFill("solid", fgColor=Color.BLUE.value)
    YELLOW = PatternFill("solid", fgColor=Color.YELLOW.value)
    PINK = PatternFill("solid", fgColor=Color.PINK.value)
    LIGHT_BLUE = PatternFill("solid", fgColor=Color.LIGHT_BLUE.value)
    GRAY = PatternFill("solid", fgColor=Color.GRAY.value)
    ORANGE = PatternFill("solid", fgColor=Color.ORANGE.value)


class HAlign(Enum):
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"
    JUSTIFY = "justify"


class VAlign(Enum):
    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"
    JUSTIFY = "justify"


default_border = Border(left=Side(style='thin', color=Color.BLACK.value),
                        right=Side(style='thin', color=Color.BLACK.value),
                        top=Side(style='thin', color=Color.BLACK.value),
                        bottom=Side(style='thin', color=Color.BLACK.value),
                        vertical=Side(style='thin', color=Color.BLACK.value),
                        horizontal=Side(style='thin', color=Color.BLACK.value))


class ExcelType(Enum):
    XLS = "xls"
    XLSX = "xlsx"
    UNKNOWN = "unknown"


class MyBook:
    def __init__(self, path=None):
        """
        Load or create a workbook, if path is None will create a new xlsx else will load the excel file
        :param path:
        """
        self.pre = None
        self.path = Path(path) if path is not None else None

        if self.path is not None and not self.path.exists():
            raise IOError(f"Not exists {str(self.path)}!")

        if self.path is None:
            # create a new xlsx, only support xlsx
            self.excel_type = ExcelType.XLSX
            from openpyxl import Workbook
            self.pre = Workbook()
        else:
            # load workbook , support xlsx and xls
            if self.path.suffix in [".xlsx", ".xlsm"]:
                self.excel_type = ExcelType.XLSX
                from openpyxl import load_workbook
                self.pre = load_workbook(self.path, data_only=True)
            elif self.path.suffix == ".xls":
                self.excel_type = ExcelType.XLS
                from xlrd import open_workbook
                self.pre = open_workbook(self.path, formatting_info=True)
            else:
                raise TypeError(f"Not support file type:{self.path.suffix}")

    def sheet_by_index(self, index):
        if self.excel_type == ExcelType.XLS:
            ws = self.pre.sheet_by_index(index)
        else:
            ws = self.pre.worksheets[index]
        return MySheet(ws, self)

    def sheet_by_name(self, name):
        if self.excel_type == ExcelType.XLS:
            ws = self.pre.sheet_by_name(name)
        else:
            ws = self.pre.get_sheet_by_name(name)
        return MySheet(ws, self)

    def sheet_names(self):
        if self.excel_type == ExcelType.XLS:
            return self.pre.sheet_names()
        else:
            return self.pre.get_sheet_names()

    def sheets(self):
        if self.excel_type == ExcelType.XLS:
            ws = self.pre.sheets()
        else:
            ws = self.pre.worksheets
        return (MySheet(x, self) for x in ws)

    def add_sheet(self, name):
        if self.excel_type == ExcelType.XLS:
            raise (f"Not support {ExcelType.XLS.value}.")
        else:
            ws = self.pre.create_sheet(title=name)
        return MySheet(ws, self)

    def remove_sheet(self, name):
        if self.excel_type == ExcelType.XLS:
            raise (f"Not support {ExcelType.XLS.value}.")
        else:
            ws = self.sheet_by_name(name).pre
            self.pre.remove(ws)
        return None

    def save(self, new_file=None):
        if new_file:
            save_path = new_file
        elif self.path:
            save_path = self.path
        else:
            save_path = "temp.xlsx"
        self.pre.save(save_path)


class MySheet:
    def __init__(self, ws, parent_wb):
        self.wb = parent_wb
        self.pre = ws
        self.excel_type = self.wb.excel_type
        self.name = ws.name if self.excel_type == ExcelType.XLS else ws.title
        self.nrows = ws.nrows if self.excel_type == ExcelType.XLS else ws.max_row
        self.ncols = ws.ncols if self.excel_type == ExcelType.XLS else ws.max_column

    def get_cell(self, row, col):
        if self.excel_type == ExcelType.XLS:
            return MyCell(self.pre.cell(row, col), self, row, col)
        else:
            return MyCell(self.pre.cell(row=row + 1, column=col + 1), self, row, col)

    def gen_cells(self, min_idx=0, max_idx=None, start=0, end=None, by_col=False, flatten=True):
        """
        Get multi cells
        :param min_idx:
        :param max_idx:
        :param start:
        :param end:
        :param by_col:
        :param flatten: True, yield cell one by one otherwise yield an generator with cells
        :return:
        """
        if not by_col:
            act_max = max_idx if max_idx or max_idx == 0 else self.nrows - 1
            act_end = end if end or end == 0 else self.ncols - 1
        else:
            act_max = max_idx if max_idx or max_idx == 0 else self.ncols - 1
            act_end = end if end or end == 0 else self.nrows - 1
        idx_list = range(min_idx, act_max + 1)
        if self.excel_type == ExcelType.XLS:
            if not by_col:
                for act_idx in idx_list:
                    one_slice = (MyCell(x, self, act_idx, start + y) for y, x in
                                 enumerate(self.pre.row_slice(act_idx, start, act_end + 1)))
                    if not flatten:
                        yield one_slice
                    else:
                        yield from one_slice
            else:
                for act_idx in idx_list:
                    one_slice = (MyCell(x, self, start + y, act_idx) for y, x in
                                 enumerate(self.pre.col_slice(act_idx, start, act_end + 1)))
                    if not flatten:
                        yield one_slice
                    else:
                        yield from one_slice
        else:
            if not by_col:
                cells = self.pre.iter_rows(min_row=idx_list[0] + 1, max_row=idx_list[-1] + 1, min_col=start + 1,
                                           max_col=act_end + 1)
            else:
                cells = self.pre.iter_cols(min_row=start + 1, max_row=act_end + 1, min_col=idx_list[0] + 1,
                                           max_col=idx_list[-1] + 1)
            for one_row_col in cells:  # generate one row/column
                one_slice = (MyCell(x, self) for x in one_row_col)
                if not flatten:
                    yield one_slice
                else:
                    yield from one_slice

    def set_value(self, data, start_idx=(0, 0), colour=None, fill=None, border=False, halign=HAlign.JUSTIFY,
                  valign=VAlign.JUSTIFY):
        """
        set values for cells
        :param data:
        :param start_idx:
        :param colour:
        :param fill:
        :param border:
        :param halign:
        :param valign:
        :return:
        """
        ''' if by_col = False , first dimi write to one row '''
        for i, one_row in enumerate(data):
            if isinstance(one_row, list) or isinstance(one_row, tuple):
                for j, one in enumerate(one_row):
                    cell_obj = self.pre.cell(row=i + 1 + start_idx[0], column=j + 1 + start_idx[1])
                    cell_obj.value = one
                    cell_obj.alignment = Alignment(horizontal=halign.value, vertical=valign.value)
                    if colour:
                        tmp_font = copy(cell_obj.font)
                        tmp_font.color = colour.value if isinstance(colour, Color) else colour
                        cell_obj.font = tmp_font
                    if fill:
                        cell_obj.fill = fill.value
                    if border:
                        cell_obj.border = default_border

            else:
                cell_obj = self.pre.cell(row=i + 1 + start_idx[0], column=1 + start_idx[1])
                cell_obj.value = one_row
                cell_obj.alignment = Alignment(horizontal=halign.value, vertical=valign.value)
                if colour:
                    tmp_font = copy(cell_obj.font)
                    tmp_font.color = colour.value if isinstance(colour, Color) else colour
                    cell_obj.font = tmp_font
                if fill:
                    cell_obj.fill = fill.value
                if border:
                    cell_obj.border = default_border

    def set_style_auto_width(self, cols=None):
        """
        set column width
        :param cols: specify cols like [0,1,2] , None means all columns
        :return:
        """
        for i, one_col in enumerate(self.pre.columns):
            col_letter = get_column_letter(i + 1)
            if cols and i not in cols:
                continue
            max_col_width = max([len(str(one_col[x].value)) for x in range(len(one_col))])
            self.pre.column_dimensions[col_letter].width = 100 if max_col_width > 100 else max_col_width + 4

    def find(self, text, start_row=0, start_col=0, whole_match=True, by_row=True, ignore_case=False, silent=False):
        """
        add by terra,used to find a cell which match the text (whole or part)
        :param text: the text need to find
        :param start_row:
        :param start_col:
        :param whole_match: whole match or part match ,default is whole match
        :param by_row: by_row or by_col to search,default is by row
        :param ignore_case: ignore case or not,default is not ignore case
        :param silent: if True, will not print warning message when not found
        :return: if found return a cell else None
        """
        if by_row:
            for rowIndex in range(start_row, self.nrows):
                for colIndex in range(start_col, self.ncols):
                    cell = self.get_cell(rowIndex, colIndex)
                    cell_val = cell.value
                    if ignore_case:
                        cell_val = cell_val.lower()
                        text = text.lower()
                    if whole_match:
                        if cell_val == text:
                            return cell
                    else:
                        if text in cell_val:
                            return cell
        else:
            for colIndex in range(start_col, self.ncols):
                for rowIndex in range(start_row, self.nrows):
                    cell = self.get_cell(rowIndex, colIndex)
                    cell_val = cell.value
                    if ignore_case:
                        cell_val = cell_val.lower()
                        text = text.lower()
                    if whole_match:
                        if cell_val == text:
                            return cell
                    else:
                        if text in cell_val:
                            return cell
        if not silent:
            print("Warning: Can not find [%s] in sheet %s！" % (text, self.name))
        return None

    def merge_cells(self, cell1, cell2, border=True):
        """
        merge cells, for openpyxl only
        :param cell1:
        :param cell2:
        :param border:
        :return:
        """
        if border:
            cell1.pre.border = default_border
            cell2.pre.border = default_border
        self.pre.merge_cells(f"{cell1.addr}:{cell2.addr}")

    def group_rows(self, row_start, row_end, outline_level=1, hidden=False):
        """
        group rows, for openpyxl only
        :param row_start: the start row number, this row will be grouped
        :param row_end:  the last row number, this row will be grouped
        :param outline_level: group outline level,1-7,if set multi level, must set larger number first
        :param hidden: whether hiddern the group
        :return:
        """
        self.pre.row_dimensions.group(row_start + 1, row_end + 1, outline_level=outline_level, hidden=hidden)


class MyCell:
    def __init__(self, cell, parent_ws, row_idx=0, col_idx=0):
        self.pre = cell
        self.ws = parent_ws
        self.merged_val = None
        self.excel_type = self.ws.excel_type
        if self.excel_type == ExcelType.XLS:
            self.row_idx, self.col_idx = row_idx, col_idx
            self.trans_from_xlrd(cell)
        else:
            self.row_idx, self.col_idx = cell.row - 1, cell.column - 1
            self.trans_from_pyxl(cell)

    def __repr__(self):
        return '{0.addr} : {0.value} {0.bcolour}'.format(self)

    @property
    def comment(self):
        if self.excel_type == ExcelType.XLSX:
            if hasattr(self.pre, "comment"):
                return self.pre.comment
            else:
                return None
        else:
            if hasattr(self.ws.pre, "cell_note_map"):
                return self.ws.pre.cell_note_map.get((self.row_idx, self.col_idx))
            else:
                return None

    def trans_from_xlrd(self, cell):
        if cell.ctype in (2, 3) and int(cell.value) == cell.value:
            cell.value = int(cell.value)
        self.value = str(cell.value).strip()
        self.ctype = self.xlrd_ctype(cell.ctype)
        cell_fmt = self.ws.wb.pre.xf_list[cell.xf_index]
        font_fmt = self.ws.wb.pre.font_list[cell_fmt.font_index]
        self.bcolour = cell_fmt.background.pattern_colour_index
        self.bcolour_rgb = self.ws.wb.pre.colour_map[self.bcolour]
        self.bcolour_hex = self.__rgb_to_hex_color(self.bcolour_rgb)
        self.fcolour = font_fmt.colour_index
        self.fcolour_rgb = self.ws.wb.pre.colour_map[self.fcolour]
        self.fcolour_hex = self.__rgb_to_hex_color(self.fcolour_rgb)
        self.strike = font_fmt.struck_out
        self.merged = False
        for (row, row_range, col, col_range) in self.ws.pre.merged_cells:
            if self.row_idx >= row and self.row_idx < row_range and self.col_idx >= col and self.col_idx < col_range:
                if self.row_idx == row and self.col_idx == col:
                    break  # change continue to break by terra 2020/10/20
                self.merged = True
                base_cell = self.ws.pre.cell(row, col)
                self.merged_val = str(base_cell.value).strip() if base_cell.value is not None else ""
                self.value = MERGED_VALUE  # speicial nam for parse
                break  # change continue to break by terra 2020/10/20
        from xlrd import cellname
        self.addr = cellname(self.row_idx, self.col_idx)

    def trans_from_pyxl(self, cell):
        self.value = str(cell.value).strip() if cell.value is not None else ""
        self.ctype = 'text'  ## default always using text in openpyxl , update later
        self.bcolour = cell.fill.bgColor.value
        self.fcolour = cell.font.color
        self.strike = cell.font.strike
        self.addr = cell.coordinate
        self.merged = False  # add by terra 2020/9/23 openpyxl
        for cell_range in self.ws.pre.merged_cells:
            if cell_range.min_row <= self.row_idx + 1 <= cell_range.max_row and cell_range.min_col <= self.col_idx + 1 <= cell_range.max_col:
                if self.row_idx + 1 == cell_range.min_row and self.col_idx + 1 == cell_range.min_col:
                    break  # change continue to break by terra 2020/10/20
                self.merged = True
                base_cell = self.ws.pre.cell(cell_range.min_row, cell_range.min_col)
                self.merged_val = str(base_cell.value).strip() if base_cell.value is not None else ""
                self.value = MERGED_VALUE
                break  # change continue to break by terra 2020/10/20

    def xlrd_ctype(self, ctype):
        return 'text'

    def set_value(self, value, colour=None):
        if self.excel_type == ExcelType.XLSX:
            self.pre.value = value
            if colour:
                self.set_style_fcolour(colour)
        else:
            raise NotImplemented("Not support for xlrd.")

    def set_style_bcolour(self, fill):
        if self.excel_type == ExcelType.XLSX:
            self.pre.fill = fill.value
        else:
            raise NotImplemented("Not support for xlrd.")

    def set_style_fcolour(self, colour):
        if self.excel_type == ExcelType.XLSX:
            tmp_font = copy(self.pre.font)
            tmp_font.color = colour
            self.pre.font = tmp_font
        else:
            raise NotImplemented("Not support for xlrd.")

    def set_style_alignment(self, halign=HAlign.JUSTIFY, valign=VAlign.JUSTIFY):
        if self.excel_type == ExcelType.XLSX:
            self.pre.alignment = Alignment(horizontal=halign.value, vertical=valign.value)
        else:
            raise NotImplemented("Not support for xlrd.")

    def set_style_border(self, pos=None, style='thin'):
        """
        :param pos: top, left, right, bottom , None means all
        :param style:
        :return:
        """
        if self.excel_type == ExcelType.XLSX:
            from openpyxl.styles import Side
            new_side = Side(border_style=style, color=Color.BLACK.value)
            tmp_border = copy(self.pre.border)
            if not pos:
                pos = ['top', 'left', 'right', 'bottom']
            for one in pos:
                assert one in ('top', 'left', 'right', 'bottom'), "border position wrong : {}".format(one)
                setattr(tmp_border, one, new_side)
            self.pre.border = tmp_border
        else:
            raise NotImplemented("Not support for xlrd.")

    def __rgb_to_hex_color(self, rgb):
        if rgb:
            r, g, b = rgb
            return f"#{'{:02X}'.format(r)}{'{:02X}'.format(g)}{'{:02X}'.format(b)}"
        else:
            return None

    def __hex_color_to_rgb(self, hex_color):

        hex_color = hex_color.strip("#")
        if len(hex_color) == 8:
            hex_color = hex_color[2:]
        return int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)


if __name__ == '__main__':
    wb = MyBook(r"C:\Users\terra\Desktop\test.xls")
    ws = wb.sheet_by_index(0)
    for one in ws.gen_cells():
        print(f"{one.addr} {one.value} {one.merged_val}")



# coding=utf-8
import enum
import os
import subprocess


class RC_CODE(enum.Enum):
    RC_PASS = 0
    RC_FAIL = 1
    RC_EXCEPTION = -1


class RunCMD():
    def __init__(self):
        self.active_info = False
        self.use_mylog = False
        self.show_cmd = True
        self.silent = False
        self.encoding = "utf-8"

    def __bytes2str(self, bytes_data):
        """
        bytes to str
        :param bytes_data:
        :return:
        """
        try:
            return bytes_data.decode(self.encoding)
        except:
            return bytes_data.decode("gbk", "ignore")

    def run_command(self, command, cwd=os.getcwd(), skip_blank_lines=False):
        if self.show_cmd:
            print(f"Run command:{command} at '{cwd}'")
        if self.active_info:
            lines = []
            p = None
            try:
                p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                     universal_newlines=False, cwd=cwd)
                while p.poll() is None:
                    line = self.__bytes2str(p.stdout.readline())
                    if skip_blank_lines and not line.strip():
                        continue
                    if not self.silent:
                        print(line.strip())
                    lines.append(line)
                try:
                    last_line = self.__bytes2str(p.stdout.readline())
                    if not self.silent:
                        print(last_line.strip())
                    lines.append(last_line)
                except:
                    pass

                if p.returncode == 0:
                    return RC_CODE.RC_PASS, lines
                else:
                    return RC_CODE.RC_FAIL, lines

            except Exception as ex:
                print(str(ex))
                lines.append(str(ex))
                return RC_CODE.RC_EXCEPTION, lines
            finally:
                try:
                    p.terminate()
                except:
                    pass
        else:
            lines = []
            p = None
            try:
                p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                     universal_newlines=False, cwd=cwd)
                out, err = p.communicate()
                lines = [x + "\n" for x in self.__bytes2str(out).split('\n')]
                if p.returncode == 0:
                    if not self.silent:
                        for line in lines:
                            print(line.strip())
                    return RC_CODE.RC_PASS, lines
                else:
                    for line in lines:
                        print(line.strip())
                    return RC_CODE.RC_FAIL, lines
            except Exception as ex:
                print(str(ex))
                lines.append(str(ex))
                return RC_CODE.RC_EXCEPTION, lines
            finally:
                try:
                    p.terminate()
                except:
                    pass



# coding=utf-8
import decimal
import datetime
import re
from decimal import Decimal


def timestamp_to_time(timestamp):
    try:
        timestamp = float(timestamp)
        return datetime.datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")
    except:
        print(f"[WARN]Can not convert [{timestamp}] to time str!")
        return timestamp


def current_time():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")


class Time():
    def __init__(self, time):
        t = str(time).lower().replace(" ", "").strip()
        s_mo = re.match(r"^([+-]?[\d.]+)s?$", t)
        ms_mo = re.match(r"^([+-]?[\d.]+)ms?$", t)
        us_mo = re.match(r"^([+-]?[\d.]+)us?$", t)
        ns_mo = re.match(r"^([+-]?[\d.]+)ns?$", t)
        ps_mo = re.match(r"^([+-]?[\d.]+)ps?$", t)
        fs_mo = re.match(r"^([+-]?[\d.]+)fs?$", t)

        if s_mo:
            self.value = Decimal(s_mo.group(1))
            self.unit = "s"
        elif ms_mo:
            self.value = Decimal(ms_mo.group(1))
            self.unit = "ms"
        elif us_mo:
            self.value = Decimal(us_mo.group(1))
            self.unit = "us"
        elif ns_mo:
            self.value = Decimal(ns_mo.group(1))
            self.unit = "ns"
        elif ps_mo:
            self.value = Decimal(ps_mo.group(1))
            self.unit = "ps"
        elif fs_mo:
            self.value = Decimal(fs_mo.group(1))
            self.unit = "fs"
        else:
            raise ValueError(f"Can not accept the format for {time}!")

    def trans(self, unit):
        units = ["s", "ms", "us", "ns", "ps", "fs"]
        if unit not in units:
            raise ValueError(f"Your unit [{unit}] is not support.")
        self_idx = units.index(self.unit)
        target_idx = units.index(unit)
        delta = target_idx - self_idx
        self.value = self.value * decimal.ExtendedContext.power(1000, delta)
        self.unit = unit
        return self

    def __repr__(self):
        return f"{self.value}{self.unit}"

    def __add__(self, other):
        other = other.trans(self.unit)
        return Time(f"{self.value + other.value}{self.unit}")

    def __sub__(self, other):
        other = other.trans(self.unit)
        return Time(f"{self.value - other.value}{self.unit}")

    def __lt__(self, other):
        other = other.trans(self.unit)
        return self.value.__lt__(other.value)

    def __le__(self, other):
        other = other.trans(self.unit)
        return self.value.__le__(other.value)

    def __gt__(self, other):
        other = other.trans(self.unit)
        return self.value.__gt__(other.value)

    def __ge__(self, other):
        other = other.trans(self.unit)
        return self.value.__ge__(other.value)

    def __eq__(self, other):
        other = other.trans(self.unit)
        return self.value.__eq__(other.value)



import re
import random
import sys


class BinValue:
    def __init__(self, origin, type, bin_t, width, isRandom=False):
        self.origin = origin
        self.type = type
        self.bin_t = bin_t
        self.binV = bin_t[2:] if bin_t else None
        self.width = int(width) if width else None
        self.isRandom = isRandom


class CommonMethodForAsic:
    @staticmethod
    def any2bin(value, width=8, info=""):
        value = str(value).strip().lower()  # it is very strange, 8'hFF have different result with 8'hff even with re.I
        value = value.replace("_", "")
        # so i add lower()
        # random

        if value == "rsvd" or value == "reserved":
            return BinValue("", None, None, None)

        # 0x12、3'b001、4'd1、10h、10b、10d
        # 4'b1100、2'o17、3'd12、8'hff
        reg_bin1 = "^(\d*)'?b([01]+)$"  # 4'b1100
        reg_bin2 = "^([01]+)b$"  # 1100b
        reg_bin3 = "^0b([01]+)$"  # 0b1100

        reg_oct1 = "^(\d*)'?o([0-7]+)$"  # 2'o17
        reg_oct2 = "^([0-7]+)o$"  # 17o
        reg_oct3 = "^0([0-7]+)$"  # 017

        reg_dec1 = "(^\d*)'?d(\d+)$"  # 3'd110
        reg_dec2 = "^([1-9]\d*)d?$"  # 110d
        reg_dec3 = "^0$"  # 0

        reg_hex1 = "^(\d*)'?h([0-9,abcdef]+)$"  # 2'h1f
        reg_hex2 = "^([0-9,abcdef]+)h$"  # 1fh
        reg_hex3 = "^0x([0-9,abcdef]+)$"  # 0x1f

        reg_rand = "^rand(om)?$"

        # random
        if re.match(reg_rand, value, re.I):
            r_value = random.randint(0, pow(2, 32) - 1)
            f = "0b{:0%sb}" % width
            val = f.format(int(r_value))
            return BinValue(value, "dec", val, None, True)
        # bin
        if re.match(reg_bin1, value, re.I):
            width = re.sub(reg_bin1, "\\1", value, re.I)
            val = re.sub(reg_bin1, "\\2", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 2))
            return BinValue(value, "bin", val, width)
        if re.match(reg_bin2, value, re.I):
            val = re.sub(reg_bin2, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 2))
            return BinValue(value, "bin", val, None)
        if re.match(reg_bin3, value, re.I):
            val = re.sub(reg_bin3, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 2))
            return BinValue(value, "bin", val, None)

        # oct
        if re.match(reg_oct1, value, re.I):
            width = re.sub(reg_oct1, "\\1", value, re.I)
            val = re.sub(reg_oct1, "\\2", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 8))
            return BinValue(value, "oct", val, width)
        if re.match(reg_oct2, value, re.I):
            val = re.sub(reg_oct2, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 8))
            return BinValue(value, "oct", val, None)
        if re.match(reg_oct3, value, re.I):
            val = re.sub(reg_oct3, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 8))
            return BinValue(value, "oct", val, None)

        # dec
        if re.match(reg_dec1, value, re.I):
            width = re.sub(reg_dec1, "\\1", value, re.I)
            val = re.sub(reg_dec1, "\\2", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val))
            return BinValue(value, "dec", val, width)
        if re.match(reg_dec2, value, re.I):
            val = re.sub(reg_dec2, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val))
            return BinValue(value, "dec", val, None)
        if re.match(reg_dec3, value, re.I):
            f = "0b{:0%sb}" % width
            val = f.format(int(0))
            return BinValue(value, "dec", val, None)

        # hex
        if re.match(reg_hex1, value, re.I):
            x = re.findall(reg_hex1, value, re.I)
            width = re.sub(reg_hex1, "\\1", value, re.I)
            val = re.sub(reg_hex1, "\\2", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 16))
            return BinValue(value, "hex", val, width)
        if re.match(reg_hex2, value, re.I):
            val = re.sub(reg_hex2, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 16))
            return BinValue(value, "hex", val, None)
        if re.match(reg_hex3, value, re.I):
            val = re.sub(reg_hex3, "\\1", value, re.I)
            f = "0b{:0%sb}" % width
            val = f.format(int(val, 16))
            return BinValue(value, "hex", val, None)
        if value != "" and value != "none":
            print(
                '[AsicCommonMethod][Error]:"%s" can not be recognized as any format, it can not be translate to bin! %s' % (
                    value, info))
            sys.exit(1)
        return BinValue(value, None, None, None, False)

    @staticmethod
    def bin2hex(value, width="", info=""):
        """
        bin translate to width'h1f format
        :param value:
        :param width:
        :return:
        """
        reg_bin1 = "^(\d*)'?b([01]+)$"  # 4'b1100
        reg_bin2 = "^([01]+)b$"  # 1100b
        reg_bin3 = "^0b([01]+)$"  # 0b1100

        value = str(value).strip()
        if re.match(reg_bin1, value, re.I):
            val = re.sub(reg_bin1, "\\2", value, re.I)
            f = "'h{:0%sx}" % width
            val = f.format(int(val, 2))
            return val
        if re.match(reg_bin2, value, re.I):
            val = re.sub(reg_bin2, "\\1", value, re.I)
            f = "'h{:0%sx}" % width
            val = f.format(int(val, 2))
            return val
        if re.match(reg_bin3, value, re.I):
            val = re.sub(reg_bin3, "\\1", value, re.I)
            f = "'h{:0%sx}" % width
            val = f.format(int(val, 2))
            return val

        if value != "" and value != "none":
            print(
                '[AsicCommonMethod][Error]:"%s" can not be recognized as any bin format, it can not be translate to hex! %s' % (
                    value, info))
            sys.exit(1)
        return value

    @staticmethod
    def any2bin_test(value, width=8, info=""):
        value = str(value).strip().lower()  # it is very strange, 8'hFF have different result with 8'hff even with re.I
        # so i add lower()
        # random
        if value == "rsvd" or value == "reserved":
            return False
        # 0x12、3'b001、4'd1、10h、10b、10d
        # 4'b1100、2'o17、3'd12、8'hff
        reg_bin1 = "^(\d*)'?b([01]+)$"  # 4'b1100
        reg_bin2 = "^([01]+)b$"  # 1100b
        reg_bin3 = "^0b([01]+)$"  # 0b1100

        reg_oct1 = "^(\d*)'?o([0-7]+)$"  # 2'o17
        reg_oct2 = "^([0-7]+)o$"  # 17o
        reg_oct3 = "^0([0-7]+)$"  # 017

        reg_dec1 = "(^\d*)'?d(\d+)$"  # 3'd110
        reg_dec2 = "^([1-9]\d*)d?$"  # 110d
        reg_dec3 = "^0$"  # 0

        reg_hex1 = "^(\d*)'?h([0-9,abcdef]+)$"  # 2'h1f
        reg_hex2 = "^([0-9,abcdef]+)h$"  # 1fh
        reg_hex3 = "^0x([0-9,abcdef]+)$"  # 0x1f

        reg_rand = "^rand(om)?$"

        if re.match(reg_rand, value):
            return True
        # bin
        if re.match(reg_bin1, value) or re.match(reg_bin2, value) or re.match(reg_bin3, value):
            return True
        # oct
        if re.match(reg_oct1, value) or re.match(reg_oct2, value) or re.match(reg_oct3, value):
            return True
        # dec
        if re.match(reg_dec1, value) or re.match(reg_dec2, value) or re.match(reg_dec3, value):
            return True
        # hex
        if re.match(reg_hex1, value) or re.match(reg_hex2, value) or re.match(reg_hex3, value):
            return True
        if value != "" and value != "none":
            return False
        return False

    @staticmethod
    def get_bit_width(val):
        return 1 if val == 0 else val.bit_length()


def test():
    test_list = ["10", "10b", "3'b001", "4'b1100", "2'o17", "017", "4'd1", "10d", "3'd12", "10h", "8'hff", "0x12"]
    print("Test Result: \n"
          "Value{:8}Type{:6}BinValue{:18}BinValue2{:21}Width{:10}\n"
          "____________________________________________________________________________________".format("", "", "", "",
                                                                                                        ""))
    for t in test_list:
        result = CommonMethodForAsic.any2bin(t)
        print("{:13}{:10}{:26}{:30}{:10}".format(result.origin, result.type, result.bin_t, result.binV,
                                                 str(result.width)))


if __name__ == '__main__':
    test()
    print(CommonMethodForAsic.bin2hex("0b1111", "2"))
    print(CommonMethodForAsic.bin2hex("1111b", "3"))
    print(CommonMethodForAsic.bin2hex("b1111", "2"))



import subprocess
import re
import os
import sys
from distutils.dist import command_re

from . import SubProcess


class File_Info:
    def __init__(self, fileLink, fileName, fileType, revision, lastUpdateTime, lastUpdateAuthor, lockStatus, lockTime,
                 lockAuthor):
        self.FileLink = fileLink
        self.FileName = fileName
        self.FileType = fileType
        self.Revision = revision
        self.LastUpdateTime = lastUpdateTime
        self.LastUpdateAuthor = lastUpdateAuthor
        self.LockStatus = lockStatus
        self.LockTime = lockTime
        self.LockAuthor = lockAuthor


class File_Log:
    def __init__(self, fileRev, author, date, description):
        self.Revision = int(fileRev.strip())
        self.Author = author
        self.Date = date
        self.Description = description


temp_comment = "temp_comment.txt"


class SVN_Helper:
    def __init__(self, repo_url, user, pw, work_copy):
        self.repo_url = repo_url
        self.work_copy = work_copy if work_copy else "."
        self.user = user
        self.pw = pw

        if user and pw:
            self.authority = "--username={0} --password={1}".format(self.user, self.pw)
        else:
            self.authority = ""

    def __bytes2str(self, bytes_data):
        """
        bytes to str
        :param bytes_data:
        :return:
        """
        try:
            return bytes_data.decode("gbk")
        except:
            return bytes_data.decode("utf-8", "ignore")

    def __run_command(self, command, cwd=None, active_info=True, silent=True, show_cmd=True):
        run_command = SubProcess.RunCMD()
        run_command.active_info = active_info
        run_command.silent = silent
        run_command.show_cmd = show_cmd
        result, msg = run_command.run_command(command, cwd=cwd)
        if result:
            return True, msg
        else:
            return False, msg

    def svn_list(self, dir_or_file=None, exit_on_error=False):
        if dir_or_file:
            command = 'svn ls "{dir_or_file}" -R {authority}'.format(dir_or_file=dir_or_file,
                                                                     authority=self.authority)
        else:
            command = 'svn ls -R {authority}'.format(authority=self.authority)
        result, output = self.__run_command(command)
        if result:
            print("List successful!")
            return output
        else:
            print("List failed!")
            if exit_on_error:
                exit(1)
            return None

    def svn_add(self, dir_or_file=None, exit_on_error=False, silent=False):
        if dir_or_file:
            command = 'svn add "{dir_or_file}" --force {authority}'.format(dir_or_file=dir_or_file,
                                                                           authority=self.authority)
        else:
            command = 'svn add * --force {authority}'.format(authority=self.authority)
        result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
        if result:
            if not silent:
                print("Add successful!")
        else:
            print("Add failed!")
        return result

    def svn_delete(self, dir_or_file, exit_on_error=False, silent=False):
        command = 'svn delete "{dir_or_file}" --force {authority}'.format(dir_or_file=dir_or_file,
                                                                          authority=self.authority)
        result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
        if result:
            if not silent:
                print("Delete successful!")
        else:
            print("Delete failed!")
            if exit_on_error:
                exit(1)
        return result

    def svn_lock(self, fileUrl, exit_on_error=False):
        command = 'svn lock "{fileUrl}" {authority}'.format(fileUrl=fileUrl, authority=self.authority)
        result, output = self.__run_command(command)
        if result:
            print("Lock successful!")
        else:
            print("Lock failed!")
            if exit_on_error:
                exit(1)
        return result

    def svn_unlock(self, fileUrl, exit_on_error=False):
        command = 'svn unlock "{fileUrl}" {authority}'.format(fileUrl=fileUrl, authority=self.authority)
        result, output = self.__run_command(command)
        if result:
            print("Unlock successful!")
        else:
            print("Unlock failed!")
            if exit_on_error:
                exit(1)
        return result

    def svn_commit(self, file="", comment="", exit_on_error=False, silent=False):
        comment_file = self.prepare_comment_file(comment)
        try:
            if file:
                command = 'svn ci "{file}" -F "{comment}" {authority}'.format(file=file, comment=comment_file,
                                                                              authority=self.authority)
            else:
                command = 'svn ci -F "{comment}" {authority}'.format(comment=comment_file, authority=self.authority)
            result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
            if result:
                if not silent:
                    print("Commit successful!")
            else:
                print("Commit failed!")
                if exit_on_error:
                    exit(1)
        finally:
            if os.path.exists(comment_file):
                os.remove(comment_file)

    def svn_commit_files(self, files, comment="", depth="", exit_on_error=False, silent=False):
        comment_file = self.prepare_comment_file(comment)
        try:
            commit_files = " ".join([f'"{x}"' for x in files])
            command = f'svn ci {depth} {commit_files} -F "{comment_file}" {self.authority}'
            result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
            if result:
                print("Commit successful!")
                for line in output:
                    if line.strip().startswith("Committed revision"):
                        commit_revision = re.sub("^Committed revision (\d+).*$", "\\1", line)
                        return commit_revision
            else:
                print("Commit failed!")
                if exit_on_error:
                    exit(1)
        finally:
            if os.path.exists(comment_file):
                os.remove(comment_file)

    def svn_log(self, fileUrl, level=100):
        logResult = []
        command = 'svn log "{fileUrl}" -l {level} {authority}'.format(fileUrl=fileUrl, level=level,
                                                                      authority=self.authority)
        result, output = self.__run_command(command)
        if result:
            str_line = ''
            for line in output:
                str_line += line.strip() + '|'
            new_lines = str_line.split('------------------------------------------------------------------------')
            for new_line in new_lines:
                new_line = new_line.strip('|')
                if new_line.strip() != "":
                    parts = new_line.split('|')
                    description = ''
                    for i in range(5, len(parts)):
                        description += parts[i] + "\n"
                    rev = re.sub("^r(\d+)$", "\\1", parts[0].strip())
                    author = parts[1].strip()
                    date = re.sub('.*?(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}).*?', "\\1", parts[2])
                    description = description.strip()
                    fileLog = File_Log(rev, author, date, description)
                    logResult.append(fileLog)
        return logResult

    def svn_remote_files(self, folderUrl=None, suffix=None):
        completeFiles = []
        folderUrl = folderUrl if folderUrl else self.repo_url
        command = f'svn ls "{folderUrl}" -R {self.authority}'
        result, output = self.__run_command(command)
        if result:
            for file in output:
                if file.strip():
                    file_extension = os.path.splitext(file)[1].lower()
                    if isinstance(suffix, list):
                        if file_extension in suffix:
                            if file.strip():
                                completeFiles.append(folderUrl.strip("/") + "/" + file)
                    else:
                        completeFiles.append(folderUrl.strip("/") + "/" + file)
        return completeFiles

    def svn_file_info(self, fileUrl):
        command = 'svn info "{fileUrl}" {authority}'.format(fileUrl=fileUrl, authority=self.authority)
        result, output = self.__run_command(command)
        fileName = os.path.basename(fileUrl)
        fileType = None
        revision = 0
        lastUpdateTime = ""
        lastUpdateAuthor = ""
        lockStatus = ""
        lockTime = ""
        lockAuthor = ""
        if result:
            for line in output:
                if line.startswith("URL:"):
                    fileUrl = line[4:].strip()
                if line.startswith("Node Kind:"):
                    fileType = line[10:].strip()
                if line.startswith("Last Changed Rev:"):
                    revision = line[18:].strip()
                if line.startswith("Last Changed Author:"):
                    lastUpdateAuthor = line[20:40].strip()
                if line.startswith("Last Changed Date:"):
                    lastUpdateTime = line[18:38].strip()
                if line.startswith("Lock Owner:"):
                    lockStatus = '<font color="blue">Locked</font>'
                    lockAuthor = line[12:].strip()
                if line.startswith("Lock Created:"):
                    lockTime = line[14:34].strip()
            fileInfo = File_Info(fileUrl, fileName, fileType, revision, lastUpdateTime, lastUpdateAuthor, lockStatus,
                                 lockTime,
                                 lockAuthor)
            return fileInfo

    def svn_checkout(self, folderUrl=None, exit_on_error=False, silent=False):
        folderUrl = folderUrl if folderUrl else self.repo_url
        actions = []
        command = f'svn co "{folderUrl}" {self.work_copy} {self.authority}'
        result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
        if result:
            if not silent:
                print("Check out repo_url:{0} succeed!".format(folderUrl))
            for line in output:
                actions.append(line.replace(self.work_copy + "\\", ""))
        else:
            print("Check out repo_url: {0} failed!".format(folderUrl))
            if exit_on_error:
                exit(1)
        return actions

    def svn_get_remote_version(self):
        revision = self.svn_file_info(self.repo_url).Revision
        print("Remote version:{0}".format(revision))
        return revision

    def svn_get_local_version(self):
        revision = self.svn_file_info(self.work_copy).Revision
        print("Local version:{0}".format(revision))
        return revision

    def svn_revert(self, dir_or_file=None, recursive=True, silent=False):
        dir_or_file = dir_or_file if dir_or_file else self.work_copy
        recursive = "-R" if recursive else ""
        if not os.path.exists(dir_or_file):
            return self.svn_update(dir_or_file)
        else:
            command = f'svn revert {recursive} "{dir_or_file}" {self.authority}'
            result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
            if result:
                if not silent:
                    print(f"Revert {dir_or_file} succeed!")
            else:
                print(f"Revert {dir_or_file} failed!")
            return result

    def svn_clenup(self, dir_path, exit_on_error=False, silent=False):
        command = f'svn cleanup "{dir_path}" {self.authority}'
        result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
        if result:
            if not silent:
                print(f"[Succeed] Cleanup [{dir_path}]")
        else:
            print("[Failed] Cleanup [{dir_path}]")
            if exit_on_error:
                exit(1)
        return result

    def svn_update(self, dir_or_file=None, more_args="", exit_on_error=False, silent=False):
        dir_or_file = dir_or_file if dir_or_file else self.work_copy
        command = f'svn update "{dir_or_file}" {more_args} {self.authority}'
        result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
        if result:
            if not silent:
                print("Update:{0} succeed!".format(dir_or_file))
        else:
            print("Update:{0} failed!".format(dir_or_file))
            if exit_on_error:
                sys.exit(1)
        return result

    def svn_update_to(self, revision, dir_or_file=None, more_args="", exit_on_error=False, silent=False):
        dir_or_file = dir_or_file if dir_or_file else self.work_copy
        local_revision = self.svn_file_info(dir_or_file).Revision
        if str(revision).strip() == str(local_revision).strip():
            print(f"Update Ignored: Self:{revision}({local_revision}) : {dir_or_file}")
            return True
        else:
            command = f'svn update "{dir_or_file}" -r {revision} {more_args} {self.authority}'
            result, output = self.__run_command(command, show_cmd=not silent, active_info=not silent, silent=silent)
            if result:
                for line in output:
                    update_p = "(?:(?:updated to)|(?:at)) revision (\d+).*"
                    if re.fullmatch(update_p, line.strip(), flags=re.I):
                        repo_revision = re.sub(update_p, "\\1", line.strip(), flags=re.I)
                        act_revision = self.svn_file_info(dir_or_file).Revision
                        if not silent:
                            print(
                                f"Update Succeed: Self:{revision}({act_revision}) Repo:{repo_revision} : {dir_or_file}")
                        return result
                print("Update Failed:{0} to revision {1} failed!".format(dir_or_file, revision))
                if exit_on_error:
                    sys.exit(1)
                return result
            else:
                print("Update Failed:{0} to revision {1} failed!".format(dir_or_file, revision))
                if exit_on_error:
                    sys.exit(1)
            return result

    def svn_diff(self, local_version, remote_version):
        diff_actions = []
        command = "svn diff -r {local_version}:{remote_version} --summarize {authority}".format(
            local_version=local_version, remote_version=remote_version, authority=self.authority)
        result, output = self.__run_command(command)
        if result:
            print("Get diff succeed!")
            diff_actions = output
        else:
            print("Get diff failed!")
        return diff_actions

    def svn_get_actions(self, first_run=False):
        new_actions = []
        if first_run:
            actions = self.svn_checkout()
        else:
            local_version = self.svn_get_local_version()
            remote_version = self.svn_get_remote_version()
            if local_version == remote_version:
                return new_actions  # 如果版本一致，则actions为空
            else:
                actions = self.svn_diff(local_version, remote_version)
        for action in actions:
            if not (action.strip().startswith("M") or action.strip().startswith("A") or action.strip().startswith("D")):
                continue
            action_act = action.strip().split()[0]
            action_path = action.strip()[1:].strip()
            new_actions.append((action_act, action_path))
        return new_actions

    def svn_status(self, path, show_cmd=True, silent=False):
        command = f"svn status {path} {self.authority}"
        result, output = self.__run_command(command, show_cmd=show_cmd, silent=silent)
        return result, output

    def svn_isLatest(self, file):
        local_file_info = self.svn_file_info(file)
        local_revision = local_file_info.Revision
        remote_revision = self.svn_file_info(local_file_info.FileLink).Revision
        if remote_revision == local_revision:
            return True
        else:
            return False

    def svn_cat(self, path, revision="HEAD", show_cmd=True, silent=False):
        command = f"svn cat {path} -r {revision} {self.authority}"
        result, output = self.__run_command(command, show_cmd=show_cmd, silent=silent)
        return result, output

    @staticmethod
    def check_file_svn_version(file):
        svn = SVN_Helper(None, None, None, None)
        result = True
        try:
            result = svn.svn_isLatest(os.path.realpath(file))
        except:
            pass
        if not result:
            print("This file is not latest,please svn update below folder at first!")
            print("%s\n" % os.path.dirname(os.path.realpath(file)))
            sys.exit(1)

    @staticmethod
    def prepare_comment_file(comment):
        with open(temp_comment, "w", encoding="utf-8") as f:
            f.write(comment)
        return os.path.realpath(temp_comment)

    @staticmethod
    def svn_checkout_by_folder_list(repo, workspace_root, folder_list, exit_on_error=False):
        if os.path.exists(workspace_root):
            print(f"Workspace:[{workspace_root}] already existed!")
            return
        svn = SVN_Helper(None, None, None, None)
        # step1 create workspace
        cmd1 = f'svn co "{repo}" "{workspace_root}" --depth=empty'
        mycmd = SubProcess.RunCMD()
        mycmd.active_info = True
        rc, msg = mycmd.run_command(cmd1, cwd=os.getcwd())
        if rc == SubProcess.RC_CODE.RC_PASS:
            folder_list = sorted(folder_list, key=lambda x: x[0])
            # update all file_folders infinity
            for folder, update_args in folder_list:
                folder_real = os.path.realpath(os.path.join(workspace_root, folder.strip("/")))
                if not SVN_Helper.update_by_levels(svn, folder_real, update_args=update_args):
                    if exit_on_error:
                        sys.exit(1)
                    return False
        else:
            if exit_on_error:
                sys.exit(1)
            return False

    @staticmethod
    def update_by_levels(svn, name, revision="HEAD", islast=True, exit_on_error=False, update_args="--depth=infinity"):
        head, tail = os.path.split(name)
        if not tail:
            head, tail = os.path.split(head)
        if head and tail and not os.path.exists(head):
            if not SVN_Helper.update_by_levels(svn, head, revision=revision, islast=False, exit_on_error=exit_on_error,
                                               update_args=update_args):
                return False
        if islast:
            cmd = f'svn up "{name}" --r {revision} {update_args}'
        else:
            cmd = f'svn up "{name}" --depth=empty'
        mycmd = SubProcess.RunCMD()
        mycmd.active_info = True
        rc, msg = mycmd.run_command(cmd)
        if rc == SubProcess.RC_CODE.RC_PASS:
            return True
        else:
            if exit_on_error:
                sys.exit(1)
            return False
